import os
import pytest
import fiona
from .conftest import requires_gpkg

example_schema = {
    'geometry': 'Point',
    'properties': [('title', 'str')],
}

example_crs = {
    'a': 6370997,
    'lon_0': -100,
    'y_0': 0,
    'no_defs': True,
    'proj': 'laea',
    'x_0': 0,
    'units': 'm',
    'b': 6370997,
    'lat_0': 45,
}

example_features = [
    {
        "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
        "properties": {"title": "One"},
    },
    {
        "geometry": {"type": "Point", "coordinates": [1.0, 2.0]},
        "properties": {"title": "Two"},
    },
    {
        "geometry": {"type": "Point", "coordinates": [3.0, 4.0]},
        "properties": {"title": "Three"},
    },
]

@requires_gpkg
def test_read_gpkg(path_coutwildrnp_gpkg):
    """
    Implicitly tests writing gpkg as the fixture will create the data source on
    first request
    """
    with fiona.open(path_coutwildrnp_gpkg, "r") as src:
        assert len(src) == 67
        feature = next(iter(src))
        assert feature["geometry"]["type"] == "Polygon"
        assert feature["properties"]["NAME"] == "Mount Naomi Wilderness"

@requires_gpkg
def test_write_gpkg(tmpdir):
    path = str(tmpdir.join('foo.gpkg'))

    with fiona.open(path, 'w',
                    driver='GPKG',
                    schema=example_schema,
                    crs=example_crs) as dst:
        dst.writerecords(example_features)

    with fiona.open(path) as src:
        assert src.schema['geometry'] == 'Point'
        assert src.crs == example_crs
        assert len(src) == 3

@requires_gpkg
def test_write_multilayer_gpkg(tmpdir):
    """
    Test that writing a second layer to an existing geopackage doesn't remove
    and existing layer for the dataset.
    """
    path = str(tmpdir.join('foo.gpkg'))

    with fiona.open(path, 'w',
                    driver='GPKG',
                    schema=example_schema,
                    layer="layer1",
                    crs=example_crs) as dst:
        dst.writerecords(example_features[0:2])

    with fiona.open(path, 'w',
                    driver='GPKG',
                    schema=example_schema,
                    layer="layer2",
                    crs=example_crs) as dst:
        dst.writerecords(example_features[2:])

    with fiona.open(path, layer="layer1") as src:
        assert src.schema['geometry'] == 'Point'
        assert src.crs == example_crs
        assert len(src) == 2

    with fiona.open(path, layer="layer2") as src:
        assert src.schema['geometry'] == 'Point'
        assert src.crs == example_crs
        assert len(src) == 1



"""Test of compound CRS crash avoidance"""

import fiona


def test_compound_crs(data):
    """Don't crash"""
    prj = data.join("coutwildrnp.prj")
    prj.write("""COMPD_CS["unknown",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]],VERT_CS["unknown",VERT_DATUM["unknown",2005],UNIT["metre",1.0,AUTHORITY["EPSG","9001"]],AXIS["Up",UP]]]""")
    with fiona.open(str(data.join("coutwildrnp.shp"))) as collection:
        assert collection.crs == {}



from fiona import crs, _crs


def test_proj_keys():
    assert len(crs.all_proj_keys) == 87
    assert 'init' in crs.all_proj_keys
    assert 'proj' in crs.all_proj_keys
    assert 'no_mayo' in crs.all_proj_keys


def test_from_string():
    # A PROJ.4 string with extra whitespace.
    val = crs.from_string(
        " +proj=longlat +ellps=WGS84 +datum=WGS84  +no_defs +foo  ")
    assert len(val.items()) == 4
    assert val['proj'] == 'longlat'
    assert val['ellps'] == 'WGS84'
    assert val['datum'] == 'WGS84'
    assert val['no_defs']
    assert 'foo' not in val


def test_from_string_utm():
    # A PROJ.4 string with extra whitespace and integer UTM zone.
    val = crs.from_string(
        " +proj=utm +zone=13 +ellps=WGS84 +foo  ")
    assert len(val.items()) == 3
    assert val['proj'] == 'utm'
    assert val['ellps'] == 'WGS84'
    assert val['zone'] == 13
    assert 'foo' not in val


def test_to_string():
    # Make a string from a mapping with a few bogus items
    val = {
        'proj': 'longlat', 'ellps': 'WGS84', 'datum': 'WGS84',
        'no_defs': True, 'foo': True, 'axis': False, 'belgium': [1, 2]}
    assert crs.to_string(
        val) == "+datum=WGS84 +ellps=WGS84 +no_defs +proj=longlat"


def test_to_string_utm():
    # Make a string from a mapping with a few bogus items
    val = {
        'proj': 'utm', 'ellps': 'WGS84', 'zone': 13,
        'no_defs': True, 'foo': True, 'axis': False, 'belgium': [1, 2]}
    assert crs.to_string(
        val) == "+ellps=WGS84 +no_defs +proj=utm +zone=13"


def test_to_string_epsg():
    val = {'init': 'epsg:4326', 'no_defs': True}
    assert crs.to_string(val) == "+init=epsg:4326 +no_defs"


def test_to_string_zeroval():
    # Make a string with some 0 values (e.g. esri:102017)
    val = {'proj': 'laea', 'lat_0': 90, 'lon_0': 0, 'x_0': 0, 'y_0': 0,
           'ellps': 'WGS84', 'datum': 'WGS84', 'units': 'm', 'no_defs': True}
    assert crs.to_string(val) == (
        "+datum=WGS84 +ellps=WGS84 +lat_0=90 +lon_0=0 +no_defs +proj=laea "
        "+units=m +x_0=0 +y_0=0")


def test_from_epsg():
    val = crs.from_epsg(4326)
    assert val['init'] == "epsg:4326"
    assert val['no_defs']


def test_from_epsg_neg():
    try:
        crs.from_epsg(-1)
    except ValueError:
        pass
    except:
        raise


def test_to_string_unicode():
    # See issue #83.
    val = crs.to_string({
        u'units': u'm',
        u'no_defs': True,
        u'datum': u'NAD83',
        u'proj': u'utm',
        u'zone': 16})
    assert 'NAD83' in val


def test_wktext():
    """Test +wktext parameter is preserved."""
    proj4 = ('+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 '
             '+x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext '
             '+no_defs')
    assert 'wktext' in crs.from_string(proj4)


def test_towgs84():
    """+towgs84 is preserved"""
    proj4 = ('+proj=lcc +lat_1=49 +lat_2=46 +lat_0=47.5 '
             '+lon_0=13.33333333333333 +x_0=400000 +y_0=400000 +ellps=bessel '
             '+towgs84=577.326,90.129,463.919,5.137,1.474,5.297,2.4232 '
             '+units=m +wktext +no_defs')
    assert 'towgs84' in crs.from_string(proj4)


def test_towgs84_wkt():
    """+towgs84 +wktext are preserved in WKT"""
    proj4 = ('+proj=lcc +lat_1=49 +lat_2=46 +lat_0=47.5 '
             '+lon_0=13.33333333333333 +x_0=400000 +y_0=400000 +ellps=bessel '
             '+towgs84=577.326,90.129,463.919,5.137,1.474,5.297,2.4232 '
             '+units=m +wktext +no_defs')
    assert 'towgs84' in _crs.crs_to_wkt(proj4)
    assert 'wktext' in _crs.crs_to_wkt(proj4)



import sys
import collections

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

if sys.version_info[0] >= 3:
    from urllib.parse import urlparse
    from collections import UserDict
    from inspect import getfullargspec as getargspec
else:
    from urlparse import urlparse
    from UserDict import UserDict
    from inspect import getargspec

if sys.version_info >= (3, 3):
    from collections.abc import Mapping
else:
    from collections import Mapping

# Users can pass in objects that subclass a few different objects
# More specifically, rasterio has a CRS() class that subclasses UserDict()
# In Python 2 UserDict() is in its own module and does not subclass Mapping()
DICT_TYPES = (dict, Mapping, UserDict)



"""Dataset paths, identifiers, and filenames"""

import re
import sys

import attr

from fiona.compat import urlparse

# Supported URI schemes and their mapping to GDAL's VSI suffix.
# TODO: extend for other cloud plaforms.
SCHEMES = {
    'ftp': 'curl',
    'gzip': 'gzip',
    'http': 'curl',
    'https': 'curl',
    's3': 's3',
    'tar': 'tar',
    'zip': 'zip',
    'file': 'file'
}

CURLSCHEMES = set([k for k, v in SCHEMES.items() if v == 'curl'])

# TODO: extend for other cloud plaforms.
REMOTESCHEMES = set([k for k, v in SCHEMES.items() if v in ('curl', 's3')])


class Path(object):
    """Base class for dataset paths"""


@attr.s(slots=True)
class ParsedPath(Path):
    """Result of parsing a dataset URI/Path

    Attributes
    ----------
    path : str
        Parsed path. Includes the hostname and query string in the case
        of a URI.
    archive : str
        Parsed archive path.
    scheme : str
        URI scheme such as "https" or "zip+s3".
    """
    path = attr.ib()
    archive = attr.ib()
    scheme = attr.ib()

    @classmethod
    def from_uri(cls, uri):
        parts = urlparse(uri)
        path = parts.path
        scheme = parts.scheme or None

        if parts.query:
            path += "?" + parts.query

        if parts.scheme and parts.netloc:
            path = parts.netloc + path

        parts = path.split('!')
        path = parts.pop() if parts else None
        archive = parts.pop() if parts else None
        return ParsedPath(path, archive, scheme)

    @property
    def name(self):
        """The parsed path's original URI"""
        if not self.scheme:
            return self.path
        elif self.archive:
            return "{}://{}!{}".format(self.scheme, self.archive, self.path)
        else:
            return "{}://{}".format(self.scheme, self.path)

    @property
    def is_remote(self):
        """Test if the path is a remote, network URI"""
        return self.scheme and self.scheme.split('+')[-1] in REMOTESCHEMES

    @property
    def is_local(self):
        """Test if the path is a local URI"""
        return not self.scheme or (self.scheme and self.scheme.split('+')[-1] not in REMOTESCHEMES)


@attr.s(slots=True)
class UnparsedPath(Path):
    """Encapsulates legacy GDAL filenames

    Attributes
    ----------
    path : str
        The legacy GDAL filename.
    """
    path = attr.ib()

    @property
    def name(self):
        """The unparsed path's original path"""
        return self.path


def parse_path(path):
    """Parse a dataset's identifier or path into its parts

    Parameters
    ----------
    path : str or path-like object
        The path to be parsed.

    Returns
    -------
    ParsedPath or UnparsedPath

    Notes
    -----
    When legacy GDAL filenames are encountered, they will be returned
    in a UnparsedPath.
    """
    if isinstance(path, Path):
        return path

    # Windows drive letters (e.g. "C:\") confuse `urlparse` as they look like
    # URL schemes
    elif sys.platform == "win32" and re.match("^[a-zA-Z]\\:", path):
        return UnparsedPath(path)

    elif path.startswith('/vsi'):
        return UnparsedPath(path)

    else:
        parts = urlparse(path)

        # if the scheme is not one of Rasterio's supported schemes, we
        # return an UnparsedPath.
        if parts.scheme and not all(p in SCHEMES for p in parts.scheme.split('+')):
            return UnparsedPath(path)

        else:
            return ParsedPath.from_uri(path)


def vsi_path(path):
    """Convert a parsed path to a GDAL VSI path

    Parameters
    ----------
    path : Path
        A ParsedPath or UnparsedPath object.

    Returns
    -------
    str
    """
    if isinstance(path, UnparsedPath):
        return path.path

    elif isinstance(path, ParsedPath):

        if not path.scheme:
            return path.path

        else:
            if path.scheme.split('+')[-1] in CURLSCHEMES:
                suffix = '{}://'.format(path.scheme.split('+')[-1])
            else:
                suffix = ''

            prefix = '/'.join('vsi{0}'.format(SCHEMES[p]) for p in path.scheme.split('+') if p != 'file')

            if prefix:
                if path.archive:
                    result = '/{}/{}{}/{}'.format(prefix, suffix, path.archive, path.path.lstrip('/'))
                else:
                    result = '/{}/{}{}'.format(prefix, suffix, path.path)
            else:
                result = path.path
            return result

    else:
        raise ValueError("path must be a ParsedPath or UnparsedPath object")



"""Test listing a datasource's layers."""

import logging
import sys
import os

import pytest

import fiona
import fiona.ogrext
from fiona.errors import DriverError


def test_single_file_private(path_coutwildrnp_shp):
    with fiona.Env():
        assert fiona.ogrext._listlayers(
            path_coutwildrnp_shp) == ['coutwildrnp']


def test_single_file(path_coutwildrnp_shp):
    assert fiona.listlayers(path_coutwildrnp_shp) == ['coutwildrnp']


def test_directory(data_dir):
    assert sorted(fiona.listlayers(data_dir)) == ['coutwildrnp', 'gre']


def test_directory_trailing_slash(data_dir):
    assert sorted(fiona.listlayers(data_dir)) == ['coutwildrnp', 'gre']


def test_zip_path(path_coutwildrnp_zip):
    assert fiona.listlayers(
        'zip://{}'.format(path_coutwildrnp_zip)) == ['coutwildrnp']


def test_zip_path_arch(path_coutwildrnp_zip):
    vfs = 'zip://{}'.format(path_coutwildrnp_zip)
    assert fiona.listlayers('/coutwildrnp.shp', vfs=vfs) == ['coutwildrnp']


def test_list_not_existing(data_dir):
    """Test underlying Cython function correctly raises"""
    path = os.path.join(data_dir, "does_not_exist.geojson")
    with pytest.raises(DriverError):
        fiona.ogrext._listlayers(path)


def test_invalid_path():
    with pytest.raises(TypeError):
        fiona.listlayers(1)


def test_invalid_vfs():
    with pytest.raises(TypeError):
        fiona.listlayers("/", vfs=1)


def test_invalid_path_ioerror():
    with pytest.raises(DriverError):
        fiona.listlayers("foobar")



"""Tests for feature objects."""

import logging
import os
import shutil
import sys
import tempfile
import pytest

import fiona
from fiona import collection
from fiona.collection import Collection
from fiona.ogrext import featureRT


class TestPointRoundTrip(object):

    def setup(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'Point', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", driver="ESRI Shapefile", schema=schema)

    def teardown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)

    def test_geometry(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert (
            sorted(g['geometry'].items()) ==
            [('coordinates', (0.0, 0.0)), ('type', 'Point')])

    def test_properties(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert g['properties']['title'] == 'foo'

    def test_none_property(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': None} }
        g = featureRT(f, self.c)
        assert g['properties']['title'] is None


class TestLineStringRoundTrip(object):

    def setup(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'LineString', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", "ESRI Shapefile", schema=schema)

    def teardown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)

    def test_geometry(self):
        f = { 'id': '1',
              'geometry': { 'type': 'LineString',
                            'coordinates': [(0.0, 0.0), (1.0, 1.0)] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert (
            sorted(g['geometry'].items()) ==
            [('coordinates', [(0.0, 0.0), (1.0, 1.0)]),
             ('type', 'LineString')])

    def test_properties(self):
        f = { 'id': '1',
              'geometry': {'type': 'Point', 'coordinates': (0.0, 0.0)},
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert g['properties']['title'] == 'foo'


class TestPolygonRoundTrip(object):

    def setup(self):
        self.tempdir = tempfile.mkdtemp()
        schema = {'geometry': 'Polygon', 'properties': {'title': 'str'}}
        self.c = Collection(os.path.join(self.tempdir, "foo.shp"),
                            "w", "ESRI Shapefile", schema=schema)

    def teardown(self):
        self.c.close()
        shutil.rmtree(self.tempdir)

    def test_geometry(self):
        f = { 'id': '1',
              'geometry': { 'type': 'Polygon',
                            'coordinates':
                                [[(0.0, 0.0),
                                  (0.0, 1.0),
                                  (1.0, 1.0),
                                  (1.0, 0.0),
                                  (0.0, 0.0)]] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert (
            sorted(g['geometry'].items()) ==
            [('coordinates', [[(0.0, 0.0),
                                  (0.0, 1.0),
                                  (1.0, 1.0),
                                  (1.0, 0.0),
                                  (0.0, 0.0)]]),
             ('type', 'Polygon')])

    def test_properties(self):
        f = { 'id': '1',
              'geometry': { 'type': 'Polygon',
                            'coordinates':
                                [[(0.0, 0.0),
                                  (0.0, 1.0),
                                  (1.0, 1.0),
                                  (1.0, 0.0),
                                  (0.0, 0.0)]] },
              'properties': {'title': u'foo'} }
        g = featureRT(f, self.c)
        assert g['properties']['title'] == 'foo'


@pytest.mark.parametrize("driver, extension", [("ESRI Shapefile", "shp"), ("GeoJSON", "geojson")])
def test_feature_null_field(tmpdir, driver, extension):
    """
    In GDAL 2.2 the behaviour of OGR_F_IsFieldSet slightly changed. Some drivers
    (e.g. GeoJSON) also require fields to be explicitly set to null.
    See GH #460.
    """
    meta = {"driver": driver, "schema": {"geometry": "Point", "properties": {"RETURN_P": "str"}}}
    filename = os.path.join(str(tmpdir), "test_null."+extension)
    with fiona.open(filename, "w", **meta) as dst:
        g = {"coordinates": [1.0, 2.0], "type": "Point"}
        feature = {"geometry": g, "properties": {"RETURN_P": None}}
        dst.write(feature)

    with fiona.open(filename, "r") as src:
        feature = next(iter(src))
        assert(feature["properties"]["RETURN_P"] is None)
